#!/usr/bin/env python
import sys, os, argparse, gzip
from collections import defaultdict
from tqdm import tqdm
import pandas as pd

__program__ = os.path.split(sys.argv[0])[-1]
__version__ = "2025.1.23"

def main(args=None):
    # Path info
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    # Path info
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, __version__, sys.version.split(" ")[0], sys.executable)
    usage = "{} -i <input.tsv> -o <output.tsv> -f <format> ".format(__program__)
    epilog = "Copyright 2024 New Atlantis Labs (jolespin@newatlantis.io)"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)
    # Pipeline
    parser.add_argument("-i","--input", type=str, default="stdin", help = "path/to/input.tsv. The results of `pyhmmsearch.py` [Default: stdin]")
    parser.add_argument("-o","--output", type=str, default="stdout", help = "path/to/output.tsv [Default: stdout]")
    parser.add_argument("-f", "--format", type=str, choices={"table", "pickle"}, help="Output format")
    parser.add_argument("--no_header",action="store_true", help="Input does not have header")
    parser.add_argument("-b", "--best_hits_only",action="store_true", help="Best hits only")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # Output
    if opts.output == "stdout":
        opts.output = sys.stdout 

    # Input
    if opts.input == "stdin":
        f_input = sys.stdin 
    else:
        if opts.input.endswith(".gz"):
            f_input = gzip.open(opts.input, "rt")
        else:
            f_input = open(opts.input, "r")

    if not opts.no_header:
        next(f_input)
    
    if opts.best_hits_only:
        output = defaultdict(dict)
        for line in tqdm(f_input, desc="Reading PyHMMSearch"):
            line = line.strip()
            if line:
                id_protein, id_hmm, threshold, score, bias, best_domain_score, best_domain_bias, evalue = line.split("\t")
                score = float(score)
                evalue = float(evalue)
                update = True
                if id_protein in output:
                    existing_score = output[id_protein]["score"]
                    if score <= existing_score:
                        update = False
                if update:
                    output[id_protein]["id"] = id_hmm
                    # output[id_protein]["name"] = definition
                    output[id_protein]["evalue"] = evalue
                    output[id_protein]["score"] = score
        df_output = pd.DataFrame(output).T
        if df_output.empty:
            df_output = pd.Dataframe(columns=["id", "evalue", "score"])
    else:
        try:
            output = defaultdict(lambda: defaultdict(list))
            for line in tqdm(f_input, desc="Reading PyHMMSearch"):
                line = line.strip()
                if line:
                    id_protein, id_hmm, threshold, score, bias, best_domain_score, best_domain_bias, evalue = line.split("\t")
                    output[id_protein]["ids"].append(id_hmm)
                    # output[id_protein]["names"].append(definition)
                    output[id_protein]["evalues"].append(float(evalue))
                    output[id_protein]["scores"].append(float(score))
            df_output = pd.DataFrame(output).T
            df_output.insert(0, "number_of_hits", df_output["ids"].map(len))
        except KeyError:
            df_output = pd.DataFrame(columns=["number_of_hits", "ids", "evalues", "scores"])
    df_output.index.name = "id_protein"

    if opts.format == "pickle":
        df_output.to_pickle(opts.output)
    else:
        df_output.to_csv(opts.output, sep="\t")

    if opts.input != "stdin":
        f_input.close()
    
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
import sys, os, glob, gzip, warnings, argparse, pickle
from collections import defaultdict
from tqdm import tqdm
from pyhmmer.plan7 import HMMFile

__program__ = os.path.split(sys.argv[0])[-1]
__version__ = "2025.1.23"

def stdin_is_empty():
    return sys.stdin.isatty()

def main(args=None):
    # Options
    # =======
    # Path info
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, __version__, sys.version.split(" ")[0], sys.executable)
    usage = "{} -d <path/to/hmm_database> -b <hmm_database.pkl.gz> -f name ".format(__program__)
    epilog = "Copyright 2024 New Atlantis Labs (jolespin@newatlantis.io)"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)
    # Pipeline
    parser_database = parser.add_argument_group('Database arguments')
    parser_database.add_argument("-l", "--hmm_list", default="stdin", type=str, help="List of filepaths to HMM files. Cannot be used with --hmm_database. [Default: stdin]")
    parser_database.add_argument("-d", "--hmm_database", type=str, help="path/to/hmm_database.  If it is a single file then it is interpreted as a multi-HMM database. If it's a directory, then it will be scanned for .hmm[.gz] files (Not case sensitive). Cannot be used with --hmm_list.")
    parser_database.add_argument("-b", "--serialized_database", required=True, type=str, help="path/to/database.pkl[.gz] is dictionary of HMM models")
    parser_database.add_argument("-f", "--hmm_marker_field", default="accession", type=str, help="HMM reference type (accession, name) [Default: accession]")

    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename
    
    # Input
    # =====
    hmm_filepaths = list()
    if opts.hmm_database:
        assert opts.hmm_list == "stdin", "Cannot provide --hmm_database and --hmm_list"
        assert stdin_is_empty(), "Cannot provide stdin if using --hmm_database"
        if os.path.isdir(opts.hmm_database):
            for extension in ["hmm", "hmm.gz", "HMM", "HMM.gz"]:
                for fp in glob.glob(os.path.join(opts.hmm_database, f"*.{extension}")):
                    hmm_filepaths.append(fp)
        else:
            hmm_filepaths.append(opts.hmm_database)
    else:
        if opts.hmm_list == "stdin":
            f_in = sys.stdin
        else:
            if opts.hmm_list.endswith(".gz"):
                f_in = gzip.open(opts.hmm_list, "rt")
            else:
                f_in = open(opts.hmm_list, "r")
            
        for line in f_in:
            fp = line.strip()
            if fp:
                hmm_filepaths.append(fp)
        if f_in != sys.stdin:
            f_in.close()
            

    # Database
    # ========
    # Load HMMs
    name_to_hmm = dict()
    for fp in hmm_filepaths:
        with HMMFile(fp) as f_hmm:
            hmms = list(f_hmm)
            for hmm in tqdm(hmms, desc=f"Loading HMMs from {fp}", total=len(hmms)):
                try:
                    name = getattr(hmm, opts.hmm_marker_field)
                except AttributeError:
                    if opts.hmm_marker_field == "name":
                        warnings.warn("HMM did not have `--hmm_marker_field name`.  Try using `--hmm_marker_field accession")
                    if opts.hmm_marker_field == "accession":
                        warnings.warn("HMM did not have `--hmm_marker_field accession`.  Try using `--hmm_marker_field name`")
                    sys.exit(1)
                name = name.decode()
                name_to_hmm[name] = hmm

    # Output
    # ======
    # Write serialized database
    print(f"Writing serialized HMMs: {opts.serialized_database}")
    if opts.serialized_database.endswith((".gz", ".pgz")):
        f_out = gzip.open(opts.serialized_database, "wb")
    else:
        f_out = open(opts.serialized_database, "wb")
    pickle.dump(name_to_hmm, f_out)
    f_out.close()


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
import sys, os, glob, gzip, warnings, argparse, pickle
from collections import defaultdict
from multiprocessing import cpu_count
from tqdm import tqdm
from pyhmmer.plan7 import HMMFile
from pyhmmer.easel import SequenceFile, TextSequence, Alphabet
from pyhmmer import hmmsearch

__program__ = os.path.split(sys.argv[0])[-1]
__version__ = "2025.1.23"

# Filter 
def filter_hmmsearch_threshold(
    hit, 
    threshold,
    score_type, 
    ):
    if score_type:
        if score_type == "domain":
            score = hit.best_domain.score
        else:
            score = hit.score
        if score >= threshold:
            evalue = hit.evalue
            return (threshold, score, evalue)
    # else:
    #     score = hit.score
    #     evalue = hit.evalue
    #     return ("", score, evalue)

def main(args=None):
    # Path info
    script_directory  =  os.path.dirname(os.path.abspath( __file__ ))
    script_filename = __program__
    # Path info
    description = """
    Running: {} v{} via Python v{} | {}""".format(__program__, __version__, sys.version.split(" ")[0], sys.executable)
    usage = "{} -i <proteins.fasta> -o <output.tsv> -d ".format(__program__)
    epilog = "Copyright 2024 Josh L. Espinoza (jolespin@newatlantis.io)"

    # Parser
    parser = argparse.ArgumentParser(description=description, usage=usage, epilog=epilog, formatter_class=argparse.RawTextHelpFormatter)

    # Pipeline
    parser_io = parser.add_argument_group('I/O arguments')
    parser_io.add_argument("-i","--proteins", type=str, default="stdin", help = "path/to/proteins.fasta. stdin does not stream and loads everything into memory. [Default: stdin]")
    parser_io.add_argument("-o","--output", type=str, default="stdout", help = "path/to/output.tsv [Default: stdout]")
    parser_io.add_argument("--no_header", action="store_true", help = "No header")
    parser_io.add_argument("--tblout", type=str, help = "path/to/output.tblout")
    parser_io.add_argument("--domtblout", type=str, help = "path/to/output.domtblout")

    parser_utility = parser.add_argument_group('Utility arguments')
    parser_utility.add_argument("-p","--n_jobs", type=int, default=1,  help = "Number of threads to use [Default: 1]")

    parser_hmmsearch = parser.add_argument_group('HMMSearch arguments')
    parser_hmmsearch.add_argument("-s", "--scores_cutoff", type=str, help="path/to/scores_cutoff.tsv[.gz] [id_hmm]<tab>[score_threshold], No header.")
    parser_hmmsearch.add_argument("-f", "--hmm_marker_field", default="accession", type=str, choices={"accession", "name"}, help="HMM reference type (accession, name) [Default: accession]")
    parser_hmmsearch.add_argument("-t", "--score_type",  default="full", type=str, choices = {"full", "domain"}, help="{full, domain} [Default: full]")
    parser_hmmsearch.add_argument("-m", "--threshold_method", type=str, default="e",choices={"gathering", "noise", "trusted", "e"},  help="Cutoff threshold method [Default:  e]")
    parser_hmmsearch.add_argument("-e","--evalue", type=float, default=10.0,  help = "E-value threshold [Default: 10.0]")

    parser_database = parser.add_argument_group('Database arguments')
    parser_database.add_argument("-d", "--hmm_database", type=str, help="path/to/database.hmm cannot be used with -b/-serialized_database.  Expects a (concatenated) HMM file and not a directory. You can build a database from a directory using `serialize_hmm_models.py`")
    parser_database.add_argument("-b", "--serialized_database", type=str, help="path/to/database.pkl cannot be used with -d/--database_directory.  Database should be pickled dictionary {name:hmm}")

    # Options
    opts = parser.parse_args()
    opts.script_directory  = script_directory
    opts.script_filename = script_filename

    # Threads
    if opts.n_jobs < 0:
        opts.n_jobs = cpu_count()

    # Database
    if opts.serialized_database:
        print("Loading serialized HMM database", file=sys.stderr)
        # Load serialized database
        if opts.serialized_database.endswith((".gz", ".pgz")):
            f = gzip.open(opts.serialized_database, "rb")
        else:
            f = open(opts.serialized_database, "rb")
        name_to_hmm = pickle.load(f)
        f.close()

    else:
        if os.path.isdir(opts.hmm_database):
            raise IsADirectoryError("--hmm_database should be (concatenated) HMM file and not a directory.  You can build a database from a directory using `serialize_hmm_models.py`")
        # Load HMMs
        name_to_hmm = dict()

        with HMMFile(opts.hmm_database) as f:
            for hmm in list(f):
                name = getattr(hmm, opts.hmm_marker_field)
                assert name is not None, "-f/--hmm_marker_field {}` returned NoneType for HMM.  Try `-f/--hmm_marker_field {}` instead".format(opts.hmm_marker_field, {"accession":"name", "name":"accession"}[opts.hmm_marker_field])
                name_to_hmm[name.decode()] = hmm

    if opts.scores_cutoff:
        name_to_threshold = dict()
        if opts.scores_cutoff.endswith(".gz"):
            f_cutoff = gzip.open(opts.scores_cutoff, "rt")
        else:
            f_cutoff = open(opts.scores_cutoff, "r")

        for line in f_cutoff:
            line = line.strip()
            if line:
                name, threshold = line.split("\t")
                if name in name_to_hmm:
                    name_to_threshold[name] = float(threshold)
        f_cutoff.close()
        
        A = set(name_to_hmm.keys())
        B = set(name_to_threshold.keys())
        assert A == B, "There are {} HMMs that do not have a score threshold".format(len(A - B))

    # Output
    if opts.output == "stdout":
        f_output = sys.stdout 
    else:
        if opts.output.endswith(".gz"):
            f_output = gzip.open(opts.output, "wt")
        else:
            f_output = open(opts.output, "w")

    if not opts.no_header:
        print("id_protein", "id_hmm", "threshold", "score", "bias", "best_domain-score", "best_domain-bias", "e-value", sep="\t", file=f_output)

    # Optional outputs
    f_tblout = None
    tblout_header = True
    if opts.tblout:
        if opts.scores_cutoff:
            raise Exception("Cannot use --tblout with -s/--scores_cutoff")
        if opts.tblout.endswith(".gz"):
            f_tblout = gzip.open(opts.tblout, "wb")
        else:
            f_tblout = open(opts.tblout, "wb")
            
    f_domtblout = None
    domtblout_header = True
    if opts.domtblout:
        if opts.scores_cutoff:
            raise Exception("Cannot use --domtblout with -s/--scores_cutoff")
        if opts.domtblout.endswith(".gz"):
            f_domtblout = gzip.open(opts.domtblout, "wb")
        else:
            f_domtblout = open(opts.domtblout, "wb")
            
    # Input
    if opts.proteins == "stdin":
        from Bio.SeqIO.FastaIO import SimpleFastaParser

        proteins = list()
        for header, seq in tqdm(SimpleFastaParser(sys.stdin), f"Parsing sequences from {sys.stdin}"):
            id = header.split(" ")[0]
            digital_sequence = TextSequence(sequence=seq, name=id.encode()).digitize(Alphabet.amino())
            proteins.append(digital_sequence)

    else:
        with SequenceFile(opts.proteins, format="fasta", digital=True) as f:
            proteins = f.read_block()#sequences=opts.sequences_per_block)


            
    # Run HMMSearch  
    params = {
        "E":opts.evalue,
    }
    if opts.threshold_method != "e":
        params["bit_cutoffs"] = opts.threshold_method
        

    if opts.scores_cutoff:
        for id_hmm, hits in tqdm(zip(name_to_hmm.keys(), hmmsearch(name_to_hmm.values(), proteins, cpus=opts.n_jobs, **params)), desc="Performing HMMSearch", total=len(name_to_hmm)):
            # hits_filtered = list()
            for hit in hits:
                if hit.included:
                    result = filter_hmmsearch_threshold(
                        hit, 
                        threshold=name_to_threshold[id_hmm], 
                        score_type=opts.score_type,
                        )
                    if result:
                        id_query = hit.name.decode()

                        threshold, score, evalue = result
                        print(
                            id_query, 
                            id_hmm, 
                            threshold, 
                            "{:0.3f}".format(hit.score), 
                            "{:0.3f}".format(hit.bias), 
                            "{:0.3f}".format(hit.best_domain.score), 
                            "{:0.3f}".format(hit.best_domain.bias), 
                            "{:0.3e}".format(evalue), 
                        sep="\t", 
                        file=f_output,
                        )
                        # hits_filtered.append(hit)
            # if f_tblout:
            #     hits_filtered.write(f_tblout, format="targets", header=True)
            # if f_domtblout:
            #     hits_filtered.write(f_domtblout, format="domains", header=True)

    else:
        if opts.threshold_method == "e":
            for id_hmm, hits in tqdm(zip(name_to_hmm.keys(), hmmsearch(name_to_hmm.values(), proteins, cpus=opts.n_jobs, **params)), desc="Performing HMMSearch", total=len(name_to_hmm)):
                if f_tblout:
                    hits.write(f_tblout, format="targets", header=tblout_header)
                    tblout_header = False
                if f_domtblout:
                    hits.write(f_domtblout, format="domains", header=domtblout_header)
                    domtblout_header = False
                for hit in hits:
                    if hit.included:
                        id_query = hit.name.decode()
                        print(
                            id_query, 
                            id_hmm, 
                            "", 
                            "{:0.3f}".format(hit.score), 
                            "{:0.3f}".format(hit.bias), 
                            "{:0.3f}".format(hit.best_domain.score), 
                            "{:0.3f}".format(hit.best_domain.bias), 
                            "{:0.3e}".format(hit.evalue), 
                        sep="\t", 
                        file=f_output,
                        )
                        
        else:
            for id_hmm, hits in tqdm(zip(name_to_hmm.keys(), hmmsearch(name_to_hmm.values(), proteins, cpus=opts.n_jobs, **params)), desc="Performing HMMSearch", total=len(name_to_hmm)):
                if f_tblout:
                    hits.write(f_tblout, format="targets", header=True)
                if f_domtblout:
                    hits.write(f_domtblout, format="domains", header=True)
                for hit in hits:
                    if hit.included:
                        id_query = hit.name.decode()
                        hmm = name_to_hmm[id_hmm]
                        threshold = getattr(hmm.cutoffs, opts.threshold_method)
                        print(
                            id_query, 
                            id_hmm, 
                            threshold, 
                            "{:0.3f}".format(hit.score), 
                            "{:0.3f}".format(hit.bias), 
                            "{:0.3f}".format(hit.best_domain.score), 
                            "{:0.3f}".format(hit.best_domain.bias), 
                            "{:0.3e}".format(hit.evalue), 
                        sep="\t", 
                        file=f_output,
                        )
                        

    # Close
    if f_output != sys.stdout:
        f_output.close()
    if f_tblout:
        f_tblout.close()
    if f_domtblout:
        f_domtblout.close()

if __name__ == "__main__":
    main(sys.argv[1:])



from setuptools import setup

# Version
# version = None
# with open("./pykofamsearch/__init__.py", "r") as f:
#     for line in f.readlines():
#         line = line.strip()
#         if line.startswith("__version__"):
#             version = line.split("=")[-1].strip().strip('"')
# assert version is not None, "Check version in pykofamsearch/__init__.py"

exec(open('pyhmmsearch/__init__.py').read())

setup(
    name='pyhmmsearch',
    version=__version__,
    description='Fast implementation of HMMSEARCH optimized for high-memory systems using PyHmmer',
    url='https://github.com/jolespin/pyhmmsearch',
    author='Josh L. Espinoza',
    author_email='jolespin@newatlantis.io, jol.espinoz@gmail.com',
    license='MIT License',
    packages=["pyhmmsearch"],
    install_requires=[
        "pyhmmer >=0.10.12",
        "pandas",
        "tqdm",
        "requests",  # Assuming API interaction may still be useful
    ],
    include_package_data=False,
    entry_points={
        'console_scripts': [
            'pyhmmsearch=pyhmmsearch.pyhmmsearch:main',   # Executes pyhmmsearch.main()
            'reformat_pyhmmsearch=pyhmmsearch.reformat_pyhmmsearch:main',  # Executes reformat_pyhmmsearch.main()
            'serialize_hmm_models=pyhmmsearch.serialize_hmm_models:main',  # Executes serialize_hmm_models.main()
        ],
    },
#      scripts=[
#          "pyhmmsearch/pyhmmsearch.py",
#          "pyhmmsearch/reformat_pyhmmsearch.py",
#          "pyhmmsearch/serialize_hmm_models.py",
#          ],
)

